@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/pool_process_query.c,v 1.59 2007/08/01 03:54:56 y-asaba Exp $
+ * $Header: /cvsroot/pgpool/pgpool/pool_process_query.c,v 1.60 2007/08/29 05:44:50 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -68,6 +68,9 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 static POOL_STATUS Execute(POOL_CONNECTION *frontend, 
 						   POOL_CONNECTION_POOL *backend);
 
+static POOL_STATUS Parse(POOL_CONNECTION *frontend, 
+						 POOL_CONNECTION_POOL *backend);
+
 #ifdef NOT_USED
 static POOL_STATUS Sync(POOL_CONNECTION *frontend, 
 						   POOL_CONNECTION_POOL *backend);
@@ -114,6 +117,11 @@ static POOL_STATUS FunctionResultResponse(POOL_CONNECTION *frontend,
 static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend, 
 										   POOL_CONNECTION_POOL *backend);
 
+static POOL_STATUS send_extended_protocol_message(POOL_CONNECTION *cp,
+												  char *kind, int len,
+												  char *string);
+static POOL_STATUS send_execute_message(POOL_CONNECTION *cp,
+										int len, char *string);
 static int synchronize(POOL_CONNECTION *cp);
 static void process_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
 static int reset_backend(POOL_CONNECTION_POOL *backend, int qcnt);
@@ -832,6 +840,7 @@ static POOL_STATUS Execute(POOL_CONNECTION *frontend,
 	int status;
 	PreparedStatement *stmt;
 	int deadlock_detected = 0;
+	int checked = 0;
 
 	/* read Execute packet */
 	if (pool_read(frontend, &len, sizeof(len)) < 0)
@@ -888,22 +897,8 @@ static POOL_STATUS Execute(POOL_CONNECTION *frontend,
 		else
 		{
 			/* forward the query to the backend */
-			pool_write(cp, "E", 1);
-			sendlen = htonl(len + 4);
-			pool_write(cp, &sendlen, sizeof(sendlen));
-			pool_write(cp, string, len);
-
-			/*
-			 * send "Flush" message so that backend notices us
-			 * the completion of the command
-			 */
-			pool_write(cp, "H", 1);
-			sendlen = htonl(4);
-
-			if (pool_write_and_flush(cp, &sendlen, sizeof(sendlen)) < 0)
-			{
+			if (send_execute_message(cp, len, string))
 				return POOL_ERROR;
-			}
 		}
 
 		if (!REPLICATION)
@@ -920,9 +915,13 @@ static POOL_STATUS Execute(POOL_CONNECTION *frontend,
 			 * If a transaction is aborted on master node, pgpool send a
 			 * error query to another nodes.
 			 */
-			deadlock_detected = detect_deadlock_error(MASTER(backend), MAJOR(backend));
-			if (deadlock_detected < 0)
-				return POOL_END;
+			if (checked == 0)
+			{
+				deadlock_detected = detect_deadlock_error(MASTER(backend), MAJOR(backend));
+				if (deadlock_detected < 0)
+					return POOL_END;
+				checked = 1;
+			}
 		}
 	}
 
@@ -963,6 +962,165 @@ static POOL_STATUS Execute(POOL_CONNECTION *frontend,
 	return POOL_CONTINUE;
 }
 
+/*
+ * Extended query protocol has to send Flush message.
+ */
+static POOL_STATUS send_extended_protocol_message(POOL_CONNECTION *cp,
+												  char *kind, int len,
+												  char *string)
+{
+	int sendlen;
+
+	/* forward the query to the backend */
+	pool_write(cp, kind, 1);
+	sendlen = htonl(len + 4);
+	pool_write(cp, &sendlen, sizeof(sendlen));
+	pool_write(cp, string, len);
+
+	/*
+	 * send "Flush" message so that backend notices us
+	 * the completion of the command
+	 */
+	pool_write(cp, "H", 1);
+	sendlen = htonl(4);
+	if (pool_write_and_flush(cp, &sendlen, sizeof(sendlen)) < 0)
+	{
+		return POOL_ERROR;
+	}
+
+	return POOL_CONTINUE;
+}
+
+static POOL_STATUS send_execute_message(POOL_CONNECTION *cp,
+										int len, char *string)
+{
+	return send_extended_protocol_message(cp, "E", len, string);
+}
+
+/*
+ * process PARSE (V3 only)
+ */
+static POOL_STATUS Parse(POOL_CONNECTION *frontend, 
+						 POOL_CONNECTION_POOL *backend)
+{
+	char kind;
+	int i;
+	int len;
+	char *string;
+	char *name, *stmt;
+	int deadlock_detected = 0;
+	int checked = 0;
+
+	/* read Parse packet */
+	if (pool_read(frontend, &len, sizeof(len)) < 0)
+		return POOL_END;
+
+	len = ntohl(len) - 4;
+	string = pool_read2(frontend, len);
+
+	pool_debug("Parse: portal name <%s>", string);
+
+	name = strdup(string);
+	if (name == NULL)
+	{
+		pool_error("Parse: malloc failed: %s", strerror(errno));
+		return POOL_END;
+	}
+
+	pending_prepared_stmt = malloc(sizeof(PreparedStatement));
+	if (pending_prepared_stmt == NULL)
+	{
+		pool_error("Parse: malloc failed: %s", strerror(errno));
+		return POOL_END;
+	}
+	pending_prepared_stmt->portal_name = NULL;
+
+	if (*string)
+	{
+		pending_function = add_prepared_list;
+		pending_prepared_stmt->statement_name = name;
+	}
+	else
+	{
+		pending_function = add_unnamed_portal;
+		pending_prepared_stmt->statement_name = NULL;
+		free(name);
+	}
+
+	/* copy prepared statement string */
+	stmt = string;
+	stmt += strlen(string) + 1;
+	pending_prepared_stmt->prepared_string = strdup(stmt);
+	if (pending_prepared_stmt->prepared_string == NULL)
+	{
+		pool_error("SimpleForwardToBackend: strdup failed: %s", strerror(errno));
+		return POOL_END;
+	}
+
+	/* forward Parse message to backends */
+	for (i = 0; i < backend->num; i++)
+	{
+		POOL_CONNECTION *cp = backend->slots[i]->con;
+		int sendlen;
+
+		if (deadlock_detected)
+		{
+			pool_write(cp, "Q", 1);
+			len = strlen(POOL_ERROR_QUERY) + 1;
+			sendlen = htonl(len + 4);
+			pool_write(cp, &sendlen, sizeof(sendlen));
+			pool_write_and_flush(cp, POOL_ERROR_QUERY, len);
+		}
+		else if (send_extended_protocol_message(cp, "P", len, string))
+			return POOL_END;
+
+		if (!REPLICATION)
+			break;
+		else if (pool_config.replication_strict)
+		{
+			pool_debug("waiting for backend completing the query");
+			if (synchronize(cp))
+				return POOL_END;
+
+			/*
+			 * We must check deadlock error because a aborted transaction
+			 * by detecting deadlock isn't same on all nodes.
+			 * If a transaction is aborted on master node, pgpool send a
+			 * error query to another nodes.
+			 */
+			if (checked == 0)
+			{
+				deadlock_detected = detect_deadlock_error(MASTER(backend), MAJOR(backend));
+				if (deadlock_detected < 0)
+					return POOL_END;
+				checked = 1;
+			}
+		}
+	}
+
+	for (;;)
+	{
+		kind = pool_read_kind(backend);
+		if (kind < 0)
+		{
+			pool_error("SimpleForwardToBackend: pool_read_kind error");
+			return POOL_ERROR;
+		}
+		SimpleForwardToFrontend(kind, frontend, backend);
+		if (pool_flush(frontend) < 0)
+			return POOL_ERROR;
+
+		/*
+		 * If warning or log messages are received, we must read
+		 * one message from backend.
+		 */
+		if (kind != 'N') /* Notice Message */
+			break;
+	}
+
+	return POOL_CONTINUE;
+}
+
 #ifdef NOT_USED
 /*
  * process Sync (V3 only)
@@ -1976,7 +2134,11 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 */
 		case 'E':
 			status = Execute(frontend, backend);
-		break;
+			break;
+
+		case 'P':
+			status = Parse(frontend, backend);
+			break;
 
 		default:
 			if (MAJOR(backend) == PROTO_MAJOR_V3)
@@ -2792,48 +2954,7 @@ POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CO
 		if (pool_write(SECONDARY(backend), p, len))
 			return POOL_END;
 
-	if (kind == 'P') /* Parse message? */
-	{
-		char *stmt;
-		
-		name = strdup(p);
-		if (name == NULL)
-		{
-			pool_error("SimpleForwardToBackend: malloc failed: %s", strerror(errno));
-			return POOL_END;
-		}
-
-		pending_prepared_stmt = malloc(sizeof(PreparedStatement));
-		if (pending_prepared_stmt == NULL)
-		{
-			pool_error("SimpleForwardToBackend: malloc failed: %s", strerror(errno));
-			return POOL_END;
-		}
-		pending_prepared_stmt->portal_name = NULL;
-
-		if (*p)
-		{
-			pending_function = add_prepared_list;
-			pending_prepared_stmt->statement_name = name;
-		}
-		else
-		{
-			pending_function = add_unnamed_portal;
-			pending_prepared_stmt->statement_name = NULL;
-			free(name);
-		}
-
-		/* copy prepared statement string */
-		stmt = p;
-		stmt += strlen(p) + 1;
-		pending_prepared_stmt->prepared_string = strdup(stmt);
-		if (pending_prepared_stmt->prepared_string == NULL)
-		{
-			pool_error("SimpleForwardToBackend: strdup failed: %s", strerror(errno));
-			return POOL_END;
-		}
-	}
-	else if (kind == 'B') /* Bind message? */
+	if (kind == 'B') /* Bind message? */
 	{
 		char *stmt_name, *portal_name;
 		PreparedStatement *stmt;
@@ -2893,7 +3014,7 @@ POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CO
 		pending_prepared_stmt->prepared_string = NULL;
 	}
 
-	if (kind == 'P' || kind == 'B' || kind == 'D' || kind == 'C')
+	if (kind == 'B' || kind == 'D' || kind == 'C')
 	{
 		int i;
 		int kind1;
